; My very fourth AVR project - bootloader for ATtiny2313 v1.0
; Compatible with AVRProg-style bootloader. For more information
; visit http://www.media.mit.edu/~ladyada/techproj/Atmex/
; (for latest software & firmware)
;
; Copyright (C) 2004
;
;
; Adaptations and fixes by Marco Carnut on May-2007 to better fit with the
; KTimer Project

;This program is free software; you can redistribute it and/or modify it under
;the terms of the GNU General Public License as published by the Free Software
;Foundation; either version 2 of the License, or (at your option) any later version.

;This program is distributed in the hope that it will be useful, but WITHOUT
;ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS
;FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

;You should have received a copy of the GNU General Public License along with
;this program; if not, write to the Free Software Foundation, Inc., 59 Temple
;Place, Suite 330, Boston, MA 02111-1307 USA


.include "tn2313def.inc" 
.equ	FREQ   	= 	8000000    	; in Hz
.equ	WAITSEC =	3          	; in seconds
.equ	BAUD	=	19200		; Baud rate
.equ	UBR		=	FREQ / (16 * BAUD) - 1

;**************** constants

.equ	DevType	=	0x5E	; Device type for ATtiny2313
.equ	HW_VER	=	0x02
.equ	SW_MAJ_VER	=	0x01
.equ	SW_MIN_VER	=	0x0f

;.equ	PAGESIZE=	0x10	; 16 words - 32 bytes
.equ	SB1		=	0x0A
.equ	SB2		=	0x91
.equ	SB3		=	0x1e

.EQU	MYBOOTSTART = 0x340
;**************** Registers
.def	temp3   =   R21
.def	temp	=	R24	; Temporary register
.def	temp2	=	R25
.def	Delay	=	R17
.def	Delay2	=	R18
.def	char	=	R20		; UART character
.def	DataL	=	R22
.def	DataH	=	R23
.def	AddrL	=	R26     ; also known as register X
.def	AddrH 	= 	R27

.CSEG
.ORG MYBOOTSTART
BOOTLOADER:	
	ldi		R24, low(RAMEND)
	out		SPL, R24
    
	; Setup UART 19.2Kbps 
	ldi		temp, UBR
	out		UBRR, temp
	ldi		temp, (1<<RXEN)|(1<<TXEN)
	out		UCSRB, temp

L_BOOTLOADER:
   	ldi ZL,low( 2*GREETING)			; load the start-address of the startup message 
	ldi ZH,high(2*GREETING)			; into the Z-Register

str_g_output: 
	lpm char,Z+						; read one character of the string and increment 
									; string pointer (the Z-register)
	rcall UART_TX                   ; send the read character via the UART
	tst char
    brne str_g_output
    
L_MAINLOOP:

	rcall	UART_RX
	cpi		char, 0x1B       ; escape char
	brne	L_AUTOINC
    ; Clear registers and jumps to the main program
    ldi     ZL,16
    ldi     ZH,0
    clr     r1
L_CLR:
    st      -Z,r1
    tst     ZL
    brne    L_CLR
    rjmp    FLASHEND

L_AUTOINC:
	cpi		char, 'a'
	brne	L_NOT_AUTOINCREMENT
	ldi		char, 'Y'
	rjmp	L_END_SEND_CHAR

L_NOT_AUTOINCREMENT:
	cpi		char, 'A'
	brne	L_NOT_WRITEADDR
	rcall	UART_RX
	mov		AddrH, char
	rcall	UART_RX
	mov		AddrL, char
	lsl		AddrL
	rol		AddrH
	rjmp	L_END_WITH_CR

L_NOT_WRITEADDR:
	cpi		char, 'c'
	brne	L_NOT_WRITEMEM_LOW
	rcall	UART_RX
	mov		DataL, char
	rjmp	L_END_WITH_CR

L_NOT_WRITEMEM_LOW:
	cpi		char, 'C'
	brne	L_NOT_WRITEMEM_HIGH
	rcall	UART_RX
	mov		DataH, char
	movw	ZL, AddrL
	movw	R0, DataL
	ldi		temp, 0x1
	out		SPMCSR, temp
	spm
	adiw	AddrL, 2
	rjmp	L_END_WITH_CR

L_NOT_WRITEMEM_HIGH:
	cpi		char, 'e'
	brne	L_NOT_CHIPERASE
	; for (addr=1; addr < MYBOOTSTART; addr+= (2*PAGESIZE))
	ldi		AddrL, 1
	clr		AddrH
	rjmp	L_ERASECHIP_TEST
L_ERASECHIP:
	movw	ZL, AddrL
	ldi		temp, 0x3
	out		SPMCSR, temp
	spm
	subi	AddrL, low(-2*PAGESIZE)
	sbci	AddrH, high(-2*PAGESIZE)
L_ERASECHIP_TEST:
	ldi		temp, low(2*MYBOOTSTART)
	ldi		temp2, high(2*MYBOOTSTART)
	cp		AddrL, temp
	cpc		AddrH, temp2
	brlo	L_ERASECHIP
	rjmp	L_END_WITH_CR

L_NOT_CHIPERASE:
	cpi		char, 'm'
	brne	L_NOT_WRITEPAGE
	movw	ZL, AddrL
	ldi		temp, 0x05
	out		SPMCSR, temp
	spm
	nop
	rjmp	L_END_WITH_CR

L_NOT_WRITEPAGE:
	cpi		char, 'p'
	brne	L_NOT_PROGTYPE
	ldi		char, 'S'
	rjmp	L_END_SEND_CHAR

L_NOT_PROGTYPE:
	cpi		char, 'R'
	brne	L_NOT_READPROGMEM
	movw	ZL, AddrL
	lpm		temp, Z+
	lpm		char, Z+
	rcall	UART_TX
	movw	AddrL, ZL
	mov		char, temp
	rjmp	L_END_SEND_CHAR

L_NOT_READPROGMEM:
	cpi		char, 'D'
	brne	L_NOT_WRITEEEPROM
	rcall	UART_RX
	out		EEDR, char
	ldi		temp, 0x6
	rcall	EEPROMtalk
	rjmp	L_END_WITH_CR

L_NOT_WRITEEEPROM:
	cpi		char, 'd'
	brne	L_NOT_READEEPROM
	ldi		temp, 0x1
	rcall	EEPROMtalk
	rcall	L_END_SEND_CHAR
	
L_NOT_READEEPROM:
	cpi		char, 'F'
	breq	L_ELSE

L_NOT_READFUSE:
	cpi		char, 'r'
	breq    L_ELSE

L_NOT_READLOCK:
	cpi		char, 'N'
    breq    L_ELSE

L_NOT_READFUSE_HIGH:
	cpi		char, 't'
	brne	L_NOT_REQSUPPORTED
	ldi		char, DEVTYPE
	rcall	UART_TX
	clr		char
	rjmp	L_END_SEND_CHAR

L_NOT_REQSUPPORTED:
	cpi		char, 'l'
	breq	L_WRITELOCK
	cpi		char, 'x'
	breq	L_WRITELOCK
	cpi		char, 'y'
	breq	L_WRITELOCK
	cpi		char, 'T'
	brne	L_NOT_SETDEVTYPE
	rcall	UART_RX
L_WRITELOCK:             
    rjmp    L_END_WITH_CR

L_NOT_SETDEVTYPE:
	cpi		char, 'S'
	brne	L_NOT_SOFTID

	ldi		ZL, low(2*SOFT_ID)
	ldi		ZH, high(2*SOFT_ID)
L_SOFTID:
	lpm		char, Z+
	tst		char
	breq	L_END
	rcall	UART_TX
	rjmp	L_SOFTID

L_NOT_SOFTID:
	cpi		char, 'V'
    breq    L_SOFTVER
    cpi     char, 'v'
	brne	L_NOT_SOFTVER
L_SOFTVER:
	ldi		char, '1'
	rcall	UART_TX
	ldi		char, '2'
	rjmp	L_END_SEND_CHAR	

L_NOT_SOFTVER:
	cpi		char, 's'
	brne	L_NOT_SENDSB
	ldi		char, SB1
	rcall	UART_TX
	ldi		char, SB2
	rcall	UART_TX
	ldi		char, SB3
	rjmp	L_END_SEND_CHAR

L_NOT_SENDSB:
	cpi		char, 'P'
	breq	L_END_WITH_CR
	cpi		char, 'L'
	breq	L_END_WITH_CR
L_ELSE:
	ldi		char, '?'
	rjmp	L_END_SEND_CHAR
L_END_WITH_CR:
	ldi		char, 13
L_END_SEND_CHAR:
	rcall	UART_TX
L_END:
	rjmp	L_MAINLOOP

EEPROMtalk:
	out		EEAR, AddrL
	adiw	AddrL,1
	sbrc	temp, 1
	sbi		EECR, EEMPE
	out		EECR, temp
L_EEPROM:
	sbic	EECR, EEPE
	rjmp	L_EEPROM
	in		char, EEDR
	ret

; reads a byte in from the UART and put it in register "CHAR"
UART_RX:
	sbis	UCSRA, RXC
	rjmp	UART_RX
	in		char, UDR
	ret

; takes a byte in register "CHAR" and sends it on the UART
UART_TX:
	sbis	UCSRA, UDRE
	rjmp	UART_TX
	out		UDR, char
	ret

SOFT_ID: .DB	"AVR2313", 0

GREETING: .DB 13,10,10,"AVR910 mode, ESC quits",13,10,0